%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXNK:FAST-GC-1.SL
% Description:  1-Space Compacting Garbage Collector for MC68000 PSL
% Author:       Sam Sands
% Created:      23-August-1983
% Modified:     31-May-84 07:54:40 (Brian Beach)
% Mode:         Lisp
% Package:
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1984, Hewlett-Packard Company, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
%  1-Sep-88 (Julian Padget)
%  Changed system fluid variables to global.
% 29-May-84 15:55:51 (Bob Shaw)
%  Added gcrunning* to block recursive calls to the garbage collector.
%  Recursive calls result from the GC calling arbitrary user functions
%  which might require memory allocation.
% 16-May-84 15:44:08 (Jim Ambras/CRC)
%  Changed hardcoded values of symboltable and hashtable sizes to wconst
%  references.
% 16-Apr-84 14:18:51 (Brian Beach)
%  Added definition of %CHIPMUNK-KLUDGE.  Was previously defined in HP-ASM.
%  Removed uses of short brach instruction to internal entry points (e.g.
%  (BGT!.S (INTERNALENTRY foo)) due to assembler malfunction.
% 02-Mar-84 09:17:57 (Jim Ambras/CRC)
%  Corrected file header.
% 06-Jan-84 08:59:13 (Tim Tillson)
%    Un-rlisp'ed Sam's D register version
%  2-Nov-83 Sam Sands
%    Converted for D register model
%  Adapted from the 1-SPACE-GC.RED
%    which was last modified on 29-Apr-83 by Sam Sands and Gwyn Osnos
%    (Algorithm is similar, but NOT the same.)
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%               The Fast 1-space Garbage Collector
%
%  The major passes of this collector have been coded in assembly language
%  for speed. This is essentially a 3 pass garbage collector:
%
%   1.  Mark
%   2.  Compact Heap
%   3.  Relocate Pointers
%
%  This collector uses an auxiliary array, called GCArray, which is 1/8th the
%  size of the PSL heap.
%
%
%  Mark Phase                                (procedure MarkFromEachBase)
%
%  During this pass all lisp items which are still accessible are found and
%  marked.  Any items in the heap which remain unmarked at the end of this pass
%  are considered garbage and the space for these is recovered in pass 2.
%
%  Symbol table entries (refered to as ID's) are also marked if they are
%  referenced.  At the end of the mark phase, procedure MakeIDFreeList links
%  all the unmarked symbol table entries into a list so that they can be
%  reused.
%
%  Items are considered accessible if they can be reached from
%    1) The execution stack            (frame variables)
%    2) The binding stack              (old values of shallow bound variables)
%    3) Any of the first 129 symbols   (the ASCII characters, plus NIL)
%    4) The hash table for symbols     ('interned' identifiers)
%        or
%    5) Any item already marked as accessible
%
%  The items are traced in depth-first order from the bases (1 thru 4 above).
%  This could be accomplished by recursive procedure calls or by a pointer
%  reversal algorithm, but this collector uses GCArray as storage space for
%  a stack of descriptors for partially traced vectors.  All traceable items
%  are treated as vectors.  A pair is a vector of length 2, while a symbol
%  table entry is 2 or 3 vectors of length 1 (name, property, and value cells).
%  When an untraced vector is encountered, the address and size of the
%  remaining portion of the vector currently being traced is pushed onto the
%  stack in GCArray.  The addresses and sizes of the execution stack and
%  binding stack are pushed onto the GCArray stack as if they were vectors.
%  The trace algorithm simply loops until it has exhausted the list of
%  vectors on the stack.
%
%
%  Compact Phase                             (procedure CompactHeap)
%
%  In this pass the PSL heap is scanned linearly and each object in it is
%  examined to see whether it is marked.  (The marks are cleared during this
%  pass.)  The marked items are moved down to be contiguous.  A map is
%  constructed to record how far each object was moved.
%
%  There are actually two phases:
%    A) Scan the stable area   (that part of the heap which won't be moved)
%         i) clear marks
%        ii) stop when 1st garbage item is encountered
%    B) Compact the active area  (the part which must be made contiguous)
%         i) count garbage bytes
%        ii) construct the relocation table (map)
%       iii) clear marks
%        iv) move non-garbage items down to be contiguous
%
%  If no garbage is found, OR if no more non-garbage is found after the
%  first garbage item, then no objects are moved.  In this case the
%  garbage collection is terminated early and pass 3 is not done.
%
%
%  The Relocation Table or Map               ( GCArray )
%
%  This data structure is built during pass 2 (compaction) to be used during
%  pass 3 (pointer relocation).  Its purpose is to show how far each
%  non-garbage item was moved, so that all pointers to them can be adjusted.
%
%  There is one 32-bit entry in the table for every eight 32-bit items in
%  the active area of the heap.  The correspondence begins at the first
%  garbage item (FirstHolePtr is the address of this item).
%
%  The 8 items in the heap which correspond to one of these table entries is
%  called a segment.  (A segment is 32 bytes long, since an item is 4 bytes.)
%  Each item within the segment will have a 'segment offset' of from 0 to 7
%  from the beginning of the segment (actually 0, 4, ... , 28 bytes).
%
%  The entries in the table are filled in only if there is at least one lisp
%  object which begins in the corresponding segment of the heap.
%  The format of each table entry is as follows:
%
%           <--  8 bits --> <--        24 bits                          -->
%          +-+-+-+-+-+-+-+-+-----------------------------------------------+
%          | garbage map   |      displacement                             |
%          +-+-+-+-+-+-+-+-+-----------------------------------------------+
%
%  The displacement field indicates how far the heap objects starting in the
%  corresponding segment were moved down during pass 2.  Normally, all of the
%  objects in a given segment get moved the same distance; this is indicated
%  by leaving the garbage map field set to 0.  The only time that not all of
%  objects in a segment move the same distance is when there are garbage items
%  completely within the segment, separating non-garbage items starting in the
%  same segment.  This situation is recorded in the table entry by setting bits
%  in the garbage map field corresponding to the items within the segment
%  which are garbage.  (Direct observation of the table after a garbage collect
%  reveal that this case is actually very rare!)   If bits are set in the
%  garbage map,  then the displacement recorded is for the object with the
%  highest segment offset within the segment.
%
%    Example:  A segment contains a non-garbage pair at segment offset 2
%      and a vector starting at segment offset 6.  Everything else in the
%      segment was garbage.   The garbage bit map looks like this:
%
%                           segment offset:    7 6 5 4 3 2 1 0
%                                             +-+-+-+-+-+-+-+-+
%                           garbage map:      |0 0 1 1 0 0 0 0|
%                                             +-+-+-+-+-+-+-+-+
%                         the vector starts here ^
%                             two items of garbage ^ ^
%                                   the pair starts here ^
%                             no need to mark this garbage ^ ^
%      The displacement recorded is for the VECTOR, not the PAIR.
%
%
%
%
%  Relocate Pointers Phase                   (procedure UpdateEachBase)
%
%  In this pass, all pointers into the heap are found and are modified to
%  reflect the movement of some of the objects in the heap.  Actually,  only
%  those pointers which point into the active area (between FirstHolePtr and
%  HeapLast) are considered, since nothing moved anywhere else.  Pointers can
%  found:
%     1) in the excution stack
%     2) in the binding stack
%     3) in the name, property, and value cells of valid symbol table entries
%     4) in the heap, now contiguous, from HeapLowerBound to HeapLast
%              (previous HeapLast minus the total number of garbage bytes found)
%
%  A way to compute the relocation of the pointer:
%    (The actual algorithm coded takes some shortcuts for efficiency!)
%
%     a) subtract FirstHolePtr          => the byte offset into the active area
%     b) divide by 4  (save answer)     => the number of the item in the area
%     c) divide by 8                    => the number of the segment it's in
%     d) multiply by 4                  => the index into GCArray
%     e) add GCArrayLowerBound          => address of the desired table entry
%     f) subtract the displacement field
%          from the pointer             => approximate address of moved object
%     g) compute MODULO 8 of the
%          item number from b)          => segment offset of moved object
%     h) shift the garbage map right
%          by the segment offset        => leaves only bits to the LEFT of item
%     i) subtract 4 from pointer for
%          each bit still in map        => adjust for garbage within segment
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(fluid '(*gc        % If T, GC statistics are printed
	 gcrunning* % Indicates that GC is running (prevents recursion)
	 gctime*    % Total amount of time spent in GC
	 gcknt*     % # of GC's since system started up
	 ))


(compiletime
 (ds !%chipmunk-kludge (x)
   (inf x)
   ))

(loadtime
  (progn (setq *gc t)          % Print GC messages
         (setq gcrunning* nil) % Not currently in GC
         (setq gctime* 0)      % Initialize to zero
         (setq gcknt* 0)
         nil))

(fluid
 '(
   gcarraylowerbound  % Store major skip distances in array which
   gcarrayupperbound  % is dynamically allocated in PSL_TOP
   ))

(global  % Fluids defined elsewhere and used here.
 '(
   stacklowerbound   % Bottom of stack
   nextsymbol        % Next ID number to be allocated
   bndstklowerbound  % Bottom of binding stack
   bndstkptr         % Binding stack pointer
   heaplowerbound    % Bottom of heap
   heapupperbound    % Top of heap
   heaplast          % Next item to be allocated
   ))

(fluid   % Fluids used only here.
 '(
   ckheaplowerbound   % HeapLowerbound with $00 in high byte
   ckheaplast         % HeapLast       with $00 in high byte
   ckfirstholeptr     % FirstHolePtr   with $00 in high byte
   stackend           % Holds address of bottom of stack
   stackstart         % Holds address of top of stack
   firstholeptr       % First location moved in heap
   totalgarbagebytes  % Total amount reclaimed
   oldtime            % Milliseconds spent on last GC
   gcbaditems  
   gcbadaddress
   gcbadobject 
   ))

(setf gcbadobject (gtfixn))

(compiletime
 (flag '(compactheap gcmessage makeidfreelist markfromeachbase updateeachbase
	 gcbaditem)
       'internalfunction
	))

(de reclaim ()
  % User call to garbage collector
  (!%reclaim)
  nil
  )

(de !%reclaim ()
  % Garbage collector
  (when (not gcrunning*) 
    (let ((gcrunning* t))  % Don't allow recursion (beforegcsystemhook might cons)
      (beforegcsystemhook)))
  (setq stackend stacklowerbound)
  (setq stackstart
    (makeaddressfromstackpointer
     (adjuststackpointer (current-stack-pointer) (wminus (framesize)))))
  (when (and *gc (not gcrunning*))
    (errorprintf "*** Garbage collection starting"))
  (setq oldtime (timc))
  (setf gcknt* (wplus2 gcknt* 1))
  (setq gcbaditems 0)
  (setq ckheaplowerbound (!%chipmunk-kludge heaplowerbound))
  (setq ckheaplast (!%chipmunk-kludge heaplast))
  (markfromeachbase)  % Pass 1 - Through Symbol Table and Stacks
  (makeidfreelist)    %  minor pass through Symbol Table
  (compactheap)       % Pass 2 - Linear Traversal Through Heap
  (setq heaplast (wdifference heaplast totalgarbagebytes))
  (setq ckfirstholeptr (!%chipmunk-kludge firstholeptr))
  (when (wlessp firstholeptr heaplast)
    (updateeachbase)  % Pass 3 - Symbol Table, Stacks, and Heap
    ) 
  (setq oldtime (wdifference (timc) oldtime))
  (when (wlessp oldtime 0)
    (setq oldtime
      (wplus2 oldtime
	      (wtimes2 (wtimes2 (wtimes2 24 60) 60) 1000))))
  (setf gctime* (wplus2 gctime* oldtime))
  (when (and *gc (not gcrunning*))
    (gcmessage))
  (when (not gcrunning*) 
    (let ((gcrunning* t))  % This keeps beforegcsystemhook and aftergcsystemhook balanced
      (aftergcsystemhook)))
  nil
  )

(lap '((*entry current-stack-pointer expr 0)
       (*pop (reg 2))             % Save return address, and set stack pointer as it will 
                                  % be after the return.
       (*move (reg st) (reg 1))   % Return value is the stack pointer.
       (*push (reg 2))            % Restore the return address.
       (*exit 0)
       ))

(lap '((*entry markfromeachbase expr 0)
       
       % Mark phase of garbage collect
       %  Register usage
       %
       % D0  CKHeapLast                           A0  index hash table, symbol table
       % D1  inf, size computation, etc.          A1  current item
       % D2  mask to zap tags = 16#00FFFFFF       A2  lowest  unmarked tag = -9
       % D3  tag byte computation                 A3  address of SYMNAM array
       % D4  count of items in current vector     A4  index into current vector
       % D5  count of pending vectors             A5  index to stack of pending vectors
       % D6  highest unmarked tag = 15            A6  CKHeapLowerBound
       % D7  outer level counter                  A7 = SP

       % Setup
       (*push (reg nil))                          % Save for later          scs
       (*move (fluid symnam) (reg a3))            % Convenient base of ID NAME table
       (movea!.l (fluid gcarraylowerbound) (reg a5))      % Vector list kept in GCArray
       (moveq -1 (reg d5))                        % Vector count

       % Convenient Constants
       (movea!.w -9 (reg a2))                     % Lowest  legal unmarked tag
       (move!.l 15 (reg d6))                      % Highest legal unmarked tag
       (move!.l 16777215 (reg d2))                % To help stamp out tags
       (movea!.l (fluid ckheaplowerbound) (reg a6))       % Lower Limit of LISP Heap
       (move!.l (fluid ckheaplast) (reg d0))              % High end of Heap
                                                  % Mark everything in Execute Stack
       (movea!.l (fluid stackstart) (reg a4))
       (move!.l (fluid stackend) (reg d4))
       (sub!.l (reg a4) (reg d4))                 % Size of Execute Stack in bytes
       (asr!.l 2 (reg d4))                        % Divide by 4 to get number of items
       (subq!.l 1 (reg d4))                       % Count is from 1
       (blt!.s s1)
       (move!.l (reg d4) (postincrement (reg a5)))% push count onto vector list
       (move!.l (reg a4) (postincrement (reg a5)))% push pointer on vector list
       (addq 1 (reg d5))                          % bump count for this vector
                                                  % Mark everything in Binding Stack
  
S1     (movea!.l (fluid bndstklowerbound) (reg a4))
       (move!.l (fluid bndstkptr) (reg d4))
       (sub!.l (reg a4) (reg d4))                 % Size of Binding Stack in bytes
       (asr!.l 2 (reg d4))                        % Divide by 4 to get number of items
       (subq!.l 1 (reg d4))                       % Count is from 1
       (blt!.s b1)
       (addq 4 (reg a4))                          % Compensate for crazy indexing
       (move!.l (reg d4) (postincrement (reg a5)))% push count onto vector list
       (move!.l (reg a4) (postincrement (reg a5)))% push pointer on vector list
       (addq 1 (reg d5))                          % bump count for this vector
  
B1                                                % Now Mark all 1 char ID's, 0..127   plus NIL, 128
       (move 128 (reg d7))                        % Count 129 symbols
       (suba!.l (reg a0) (reg a0))                % Set A0 to zero (index of 1st ID)
  
LOOP0  (cmpi!.b 251 % Test to see if ID is marked
                (indexed (reg a0) (displacement (reg a3) 0)))
       (beq!.s end0)                              % Skip if marked
       (move!.l (reg a0) (reg d1))                % Offset into Symbol Table
       (bsr marksymbol)                           % Mark this symbol
  
END0   (addq 4 (reg a0))                          % Index to next symbol
       (dbf (reg d7) loop0)                       % Finally Mark all symbols accessible from the Hash Table, Obarry
       (move!.w (wconst hash-table-size) (reg d7))% Number of elements in hash table
       (*move (fluid hashtable) (reg a0))         % A0 points into Hash Table (OBArray)
  
LOOP1  (moveq 0 (reg d1))                         % D1 will hold one Hash Table entry
  
LOOP2  (move!.w (postincrement (reg a0)) % Put next Hash Table entry into D1,
                (reg d1))                         % advancing pointer into Hash Table.
       (dbgt (reg d7) loop2)                      % Fall through if Hash Table entry (D1)
                                                  % was GT 0, or count (D7) equal to -1.
                                                  % Otherwise reiterate loop
       (ble!.s end1)                              % If count was equal to -1 then done
       (add!.l (reg d1) (reg d1))                 % Multipy by 4 to index into SYMNAM
       (add!.l (reg d1) (reg d1))
       (cmpi!.b 251 % Test to see if ID is marked
                (indexed (reg d1) (displacement (reg a3) 0)))
       (dbne (reg d7) loop1)                      % Fall through if it is not marked.
                                                  % Otherwise, reiterate loop.
       (beq!.s end1)                              % If count equal to -1 then done.
       (bsr!.s marksymbol)                        % Offset into symbol table is in D1
       (dbf (reg d7) loop1)                      % Reiterate loop unless count equal to -1.
  
END1                                              %  (movea!.w 0 (reg A6))         % Lisp Convention                  scs
       (*pop (reg nil))                           % this too
       (rts)                                      % End of Mark Phase  (MarkFromEachBase)
  
MARKVECTORLOOP
       (movea!.l (predecrement (reg a5)) (reg a4))% fetch next vector ptr
       (move!.l (predecrement (reg a5)) (reg d4)) % fetch count of items
  
MARKITEMLOOP
       (move!.b (indirect (reg a4)) (reg d3))
  
UNMARKEDCAR                                       % (Enter here if you've already got the tag)
       (ext!.w (reg d3))                          % Extract TAG, sign extend to 16 bits
       (bge!.s postag)                            % Test for valid (value in range) tag
       (cmp!.w (reg a2) (reg d3))                 % Test Lower Bound
       (bge!.s case)
  
BADTAG (bra markbaditem)                          % Tag is illegal
  
POSTAG (cmp!.b (reg d6) (reg d3))                 % Test Upper Bound
       (bgt!.s badtag)
  
CASE                                              % Use a Jump Table to dispatch on the TAG stored in D3
       (add!.w (reg d3) (reg d3))                 % Multiply D3 by 2
                                                  %% The "fullword" below can be replaced when the LAP assembler
                                                  %% handles following statement correctly:
                                                  %%    (jmp (indexed (reg D3) (displacement 0) jumptable1))
       (fullword 1325084692)                      % The Jump Table
       (bra!.s markbaditem)                       % -9  Legal only in HEAP
       (bra!.s markbaditem)                       % -8  Legal only in HEAP
       (bra!.s markbaditem)                       % -7  Legal only in HEAP
       (bra!.s markbaditem)                       % -6  Legal only in HEAP
       (bra!.s markbaditem)                       % Forward      % -5  Only for that other G.C.
       (bra!.s tagbtrtag)                         % BtrTag       % -4
       (bra!.s tagunbound)                        % Unbound      % -3
       (bra!.s tagid)                             % ID           % -2
       (bra!.s tagnegint)                         % NegInt       % -1
  
JUMPTABLE1
       (bra!.s tagposint)                         % PosInt       % 0
       (bra!.s tagfixn)                           % FixN         % 1
       (bra!.s tagbign)                           % BigN         % 2
       (bra!.s tagfltn)                           % FltN         % 3
       (bra!.s tagstr)                            % Str          % 4
       (bra!.s tagbytes)                          % Bytes        % 5
       (bra!.s taghalfwords)                      % HalfWords    % 6
       (bra!.s tagwrds)                           % Wrds         % 7
       (bra!.s tagvect)                           % Vect         % 8
       (bra!.s tagpair)                           % Pair         % 9
       (bra!.s tagevect)                          % Evect        % 10
       (bra!.s markbaditem)                       % 11  Illegal Tags
       (bra!.s markbaditem)                       % 12  Illegal Tags
       (bra!.s markbaditem)                       % 13  Illegal Tags
       (bra!.s markbaditem)                       % 14  Illegal Tags
       (bra!.s tagcode)                           % Code         % 15
                                                  % If the ID referenced is NOT
  
TAGID                                             % marked, then MARK it.
       (move!.l (postincrement (reg a4)) (reg d1))% Extract INF
       (and!.l (reg d2) (reg d1))                 % use 16#00FFFFFF
       (add!.l (reg d1) (reg d1))                 % Multiply D1 by 4, to index into SYMNAM
       (add!.l (reg d1) (reg d1))
       (cmpi!.b 251 % If the SYMNAM entry for this ID is
                (indexed (reg d1) % marked FORWARD, then done ...
                         (displacement (reg a3) 0)))
       (beq!.s enditemloop)
       (subq 1 (reg d4))                          % dispense with current vector, or push it
       (bcs!.s pushid)
       (move!.l (reg d4) (postincrement (reg a5)))% save count
       (move!.l (reg a4) (postincrement (reg a5)))% save pointer
  
MARKSYMBOL                                        % (external entry point)
       (addq 1 (reg d5))                          % bump vector count
  
PUSHID (moveq 0 (reg d4))                         % construct vector of 1 value
       (*move (fluid symval) (reg a4))                      % for VALUE cell of ID
       (adda!.l (reg d1) (reg a4))                % save to do later
       (move!.l (reg d4) (postincrement (reg a5)))% push count
       (move!.l (reg a4) (postincrement (reg a5)))% push pointer
       (addq 1 (reg d5))                          %    (moveq 0 (reg D4))                       % construct vector of 1 value
       (*move (fluid symprp) (reg a4))                      % for PROPERTY cell of ID
       (adda!.l (reg d1) (reg a4))
       (cmpi!.l -33554304 (indirect (reg a4)))    % Quick test for NIL
       (beq!.s propnil)                           % save to do later
       (move!.l (reg d4) (postincrement (reg a5)))% push count
       (move!.l (reg a4) (postincrement (reg a5)))% push pointer
       (addq 1 (reg d5))
  
PROPNIL                                           %    (moveq 0 (reg D4))                       % construct vector of 1 value
       (lea (indirect (reg a3)) (reg a4))         % for SYMNAM cell of ID
       (adda!.l (reg d1) (reg a4))
       (move!.b (indirect (reg a4)) (reg d3))     % Get unmarked tag
       (move!.b 251 (indirect (reg a4)))          % Place a FORWARD mark in SYMNAM
                                                  % field to mark the ID
       (bra unmarkedcar)                          % Go mark NAM cell
  
MARKBADITEM                                       % Bad Items punished in 3rd pass of G.C.
  
TAGPOSINT                                         % Non-pointer items that do not have to be
  
TAGNEGINT                                         % examined further.
  
TAGCODE
  
TAGBTRTAG
  
TAGUNBOUND
       (addq 4 (reg a4))                          % Bump to next item in vector
  
ENDITEMLOOP
       (dbf (reg d4) markitemloop)               % Loop till end of current vector
       (dbf (reg d5) markvectorloop)             % Loop till all vectors traced
       (rts)
  
TAGSTR                                            % Pointer types
  
TAGBYTES                                          %   If the object pointed to is not in the heap,
  
TAGHALFWORDS                                      %   simply ignore it
  
TAGFIXN                                           %   Else, if the object pointed to is already marked
  
TAGBIGN                                           %   simply ignore it
  
TAGFLTN                                           %
  
TAGWRDS                                           %   Else, mark the item
  
TAGVECT                                           %   Then if it is a pair, vector, or Evector
  
TAGEVECT                                          %      recursively mark the object
  
TAGPAIR                                           %
       (move!.l (postincrement (reg a4)) (reg d1))% Extract the INF field
       (and!.l (reg d2) (reg d1))                 % using 16#00FFFFFF
       (cmp!.l (reg a6) (reg d1))                 % If the item pointed to by D1
       (blt!.s enditemloop)                       % is not within OLD heap,
       (cmp!.l (reg d0) (reg d1))                 % there is nothing to mark.
       (bge!.s enditemloop)
       (movea!.l (reg d1) (reg a1))
       (cmp!.b (indirect (reg a1)) (reg d6))      % If the item pointed to by A1 is
       (blt!.s enditemloop)                       % marked, ignore it
       (addi!.b 32 (indirect (reg a1)))           % mark the object by adding 32
       (fullword 1325084672)                      % The Jump Table
  
JUMPTABLE2
       (bra!.s enditemloop)                       % FixN
       (bra!.s enditemloop)                       % BigN
       (bra!.s enditemloop)                       % FltN
       (bra!.s enditemloop)                       % Str
       (bra!.s enditemloop)                       % Bytes
       (bra!.s enditemloop)                       % HalfWords
       (bra!.s enditemloop)                       % Wrds
       (bra!.s markvector)                        % Vect
       (bra!.s markpair)                          % Pair
                                                  % (bra!.s  MarkVector)            % Evect
  
MARKVECTOR
       (move!.l (postincrement (reg a1)) (reg d1))% Get vector header
       (addq!.l 1 (reg d1))                       % make length non-negative
       (and!.l (reg d2) (reg d1))                 % zap tag using 16#00FFFFFF
       (subq!.l 1 (reg d1))                       % restore length
       (blt!.s enditemloop)                       % ignore if <0 (null vector)
       (subq 1 (reg d4))                          % dispense with current vector, or push it
       (bcs!.s pushvect)
       (move!.l (reg d4) (postincrement (reg a5)))% save count
       (move!.l (reg a4) (postincrement (reg a5)))% save pointer
       (addq 1 (reg d5))                          % bump vector count
  
PUSHVECT
       (move!.l (reg d1) (reg d4))                % new count for this vect
       (movea!.l (reg a1) (reg a4))               % new pointer
       (bra markitemloop)
  
MARKPAIR
       (subq 1 (reg d4))                          % dispense with current vector, or push it
       (bcs!.s pushpair)
       (move!.l (reg d4) (postincrement (reg a5)))% save count
       (move!.l (reg a4) (postincrement (reg a5)))% save pointer
       (addq 1 (reg d5))                          % bump vector count
  
PUSHPAIR
       (moveq 1 (reg d4))                         % count for pair
       (movea!.l (reg a1) (reg a4))               % pointer for pair
       (move!.b (indirect (reg a4)) (reg d3))     % Fetch the tag from CAR
       (sub!.b 32 (reg d3))                       % (get rid of the mark)
       (bra unmarkedcar)))

(lap '((*entry makeidfreelist expr 0)
       % Register usage:
       % D0 - 
       % D1 -
       % D2 - 
       % D3 - String tag (tag on unmarked entries)
       % D4 - Forward tag (tag on marked entries)
       % D5
       % D6
       % D7 - Counter
       % A0 - 
       (*move (fluid symnam) (reg a3))                      % Pointer into SYMNAM Array
       (moveq 4 (reg d3))                         % STR Tag
       (moveq 251 (reg d4))                       % FORWARD Tag
       (move!.w 128 (reg d7))                     % counter
  
LOOP0  (move!.b (reg d3) (indirect (reg a3)))     % ClearIDMark (Tag := STR)
       (addq 4 (reg a3))
       (dbf (reg d7) loop0)
       (move!.w (wconst maxsymbols) (reg d7))     % Counter for remaining symbols
       (sub!.w 129 (reg d7))                      % MaxSymbols  - 129 {0..128}
       (move!.l 129 (reg d5))                     % Index into symbol table
       (lea (fluid nextsymbol) (reg a0))          % Address of Previous free symbol
  
LOOP1  (cmp!.b (indirect (reg a3)) (reg d4))      % Test for Marked ID
       (bne!.s notmarked)
       (move!.b (reg d3) (indirect (reg a3)))     % ClearIDMark (Tag := STR)
       (bra!.s bumpindex)
  
NOTMARKED
       (move!.l (reg d5) (indirect (reg a0)))
       (movea!.l (reg a3) (reg a0))               % Save address of current free symbol
  
BUMPINDEX
       (addq!.l 1 (reg d5))                       % Bump index to current symbol
       (addq 4 (reg a3))
       (dbf (reg d7) loop1)
       (move!.l 0 (indirect (reg a0)))            % Mark end of free list
       (rts)))

(lap '((*entry compactheap expr 0)
       % Assembly routine to squash out garbage
       % and build a relocation table
       %  Register usage
       %
       % D0  lowest marked tag = 23               A0  index into heap
       % D1  size computation, etc.               A1  lowest  unmarked tag = -9
       % D2  mask to zap tag, 16#00FFFFFF         A2  upper bound of heap = HeapLast
       % D3  tag byte computation                 A3  end of compacted heap
       % D4  total garbage bytes                  A4  index into relocation table
       % D5  size of relocation segment = 8       A5  highest unmarked tag = 15
       % D6  offset from beginning of segment     A6  marktag constant = 32
       % D7  size of contiguous garbage hole      A7 = SP

       % Setup
       (*push (reg nil))                          % Save for later               scs
       (movea!.l (fluid heaplowerbound) (reg a0))         % A0 points into the heap
       (movea!.l (fluid heaplast) (reg a2))               % A2 points to end of heap
       (moveq 0 (reg d4))                         % Total Garbage Bytes Found

       % Convenient Constants
       (move!.l 16777215 (reg d2))                % Mask to help stamp out tags
       (moveq 8 (reg d5))                         % Size of segment
       (movea!.w -9 (reg a1))                     % Lowest  legal tag
       (movea!.w 15 (reg a5))                     % Highest legal tag
       (moveq 23 (reg d0))                        % Lowest  legal marked tag
       (movea!.w 32 (reg a6))                     % Marktag constant
       (bra endloop1)                             % Always best to start at the bottom
  
LOOP1  (move!.b (indirect (reg a0)) (reg d3))     % get Tag byte
       (ext!.w (reg d3))                          % sign extend to 16 bits
       (sub!.w (reg a6) (reg d3))                 % Clear Mark (subtract 32)
       (bge!.s postag)                            % Test for valid tag
       (cmp!.w (reg a1) (reg d3))                 % Test Lower Bound  ( -9 )
       (bge!.s case)
       (bra!.s firstgarbage)                      % Item probably not marked
  
POSTAG (cmp!.w (reg a5) (reg d3))                 % Test Upper Bound ( 15 )
       (bgt!.s badheapitem)
  
CASE   (move!.b (reg d3) (indirect (reg a0)))     % Restore Mark in heap
       (add!.w (reg d3) (reg d3))                 % multiply by 2 to index into jump table
                                                  %% (jmp (indexed (reg D3) (displacement 0) jumptable))
       (fullword 1325084692)                      % case jump on tag of item
       (bra!.s taghbytes)                         % HBytes       % -9
       (bra!.s taghhalfwords)                     % HHalfWords   % -8
       (bra!.s taghwrds)                          % HWrds        % -7
       (bra!.s taghvect)                          % HVect        % -6
       (bra!.s badheapitem)                       % Forward      % -5     items illegal in heap
       (bra!.s badheapitem)                       % BtrTag       % -4     items illegal in heap
       (bra!.s tagunbound)                        % Unbound      % -3
       (bra!.s tagid)                             % ID           % -2
       (bra!.s tagnegint)                         % NegInt       % -1
  
JUMPTABLE
       (bra!.s tagposint)                         % PosInt       % 0
       (bra!.s tagfixn)                           % FixN         % 1
       (bra!.s tagbign)                           % BigN         % 2
       (bra!.s tagfltn)                           % FltN         % 3
       (bra!.s tagstr)                            % Str          % 4
       (bra!.s tagbytes)                          % Bytes        % 5
       (bra!.s taghalfwords)                      % HalfWords    % 6
       (bra!.s tagwrds)                           % Wrds         % 7
       (bra!.s tagvect)                           % Vect         % 8
       (bra!.s tagpair)                           % Pair         % 9
       (bra!.s tagevect)                          % Evect        % 10
       (bra!.s badheapitem)                       % Illegal Tags % 11
       (bra!.s badheapitem)                       % Illegal Tags % 12
       (bra!.s badheapitem)                       % Illegal Tags % 13
       (bra!.s badheapitem)                       % Illegal Tags % 14
       (bra!.s tagcode)                           % Code         % 15
  
TAGHBYTES
       (move!.l (indirect (reg a0)) (reg d1))
       (addq!.l 5 (reg d1))
       (addq!.l 4 (reg d1))
       (and!.l 16777212 (reg d1))
       (adda!.l (reg d1) (reg a0))
       (bra!.s endloop1)
  
TAGHHALFWORDS
       (move!.l (indirect (reg a0)) (reg d1))
       (addq!.l 4 (reg d1))
       (and!.l 16777214 (reg d1))
       (add!.l (reg d1) (reg d1))
       (adda!.l (reg d1) (reg a0))
       (bra!.s endloop1)
  
TAGHVECT
  
TAGHWRDS
       (move!.l (indirect (reg a0)) (reg d1))
       (addq!.l 2 (reg d1))
       (and!.l (reg d2) (reg d1))                 % mask with 16#00FFFFFF
       (add!.l (reg d1) (reg d1))
       (add!.l (reg d1) (reg d1))
       (adda!.l (reg d1) (reg a0))
       (bra!.s endloop1)
  
BADHEAPITEM                                       % illegal heap item assumed to be part of a pair
  
TAGUNBOUND
  
TAGID
  
TAGNEGINT
  
TAGPOSINT
  
TAGFIXN
  
TAGBIGN
  
TAGFLTN
  
TAGSTR
  
TAGBYTES                                          % all of these are assumed to be the 1st element
  
TAGHALFWORDS                                      %          of a pair
  
TAGWRDS
  
TAGVECT
  
TAGPAIR
  
TAGEVECT
  
TAGCODE(addq 8 (reg a0))                          % skip over pair
  
ENDLOOP1
       (cmpa!.l (reg a2) (reg a0))                % If A0 is already at the end of the
       (blt!.s loop1)                             %   heap, routine is done.
       (move!.l (reg a0) (fluid firstholeptr))            % No active area
       (bra done)                                 % Finish early
                                                  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
FIRSTGARBAGE
       (move!.l (reg a0) (fluid firstholeptr))            % Beginning of active area
       (movea!.l (reg a0) (reg a3))               % New HeapLast (Compacted Heap)
       (movea!.l (fluid gcarraylowerbound) (reg a4))      % Index into relocation table
       (move!.l (reg d4) (indirect (reg a4)))     % Store 1st entry in Reloc. Table
       (moveq 0 (reg d6))                         % Relative Offset into Segment
  
GTESTLOOP
       (move!.b (indirect (reg a0)) (reg d3))     % test whether item is marked
       (cmp!.b (reg d0) (reg d3))                 % Lowest marked tag is 23
       (bge copynongarbage)
  
GARBAGELOOP
       (moveq 0 (reg d7))                         % Size of contiguous garbage
  
MOREGARBAGE
       (ext!.w (reg d3))                          % sign extend to 16 bits
       (cmp!.w (reg a1) (reg d3))                 % Test Lower Bound ( -9 )
       (blt!.s gbadheapitem)                      % Item is illegal
       (add!.w (reg d3) (reg d3))                 % multiply by 2 to index into jump table
                                                  %% (jmp (indexed (reg D3) (displacement 0) jumptable))
       (fullword 1325084692)                      % case jump on tag of item
       (bra!.s tgghbytes)                         % HBytes       % -9
       (bra!.s tgghhalfwords)                     % HHalfWords   % -8
       (bra!.s tgghwrds)                          % HWrds        % -7
       (bra!.s tgghvect)                          % HVect        % -6
       (bra!.s gbadheapitem)                      % Forward      % -5     items illegal in heap
       (bra!.s gbadheapitem)                      % BtrTag       % -4     items illegal in heap
       (bra!.s tggunbound)                        % Unbound      % -3
       (bra!.s tggid)                             % ID           % -2
       (bra!.s tggnegint)                         % NegInt       % -1
  
GJUMPTABLE
       (bra!.s tggposint)                         % PosInt       % 0
       (bra!.s tggfixn)                           % FixN         % 1
       (bra!.s tggbign)                           % BigN         % 2
       (bra!.s tggfltn)                           % FltN         % 3
       (bra!.s tggstr)                            % Str          % 4
       (bra!.s tggbytes)                          % Bytes        % 5
       (bra!.s tgghalfwords)                      % HalfWords    % 6
       (bra!.s tggwrds)                           % Wrds         % 7
       (bra!.s tggvect)                           % Vect         % 8
       (bra!.s tggpair)                           % Pair         % 9
       (bra!.s tggevect)                          % Evect        % 10
       (bra!.s gbadheapitem)                      % Illegal Tags % 11
       (bra!.s gbadheapitem)                      % Illegal Tags % 12
       (bra!.s gbadheapitem)                      % Illegal Tags % 13
       (bra!.s gbadheapitem)                      % Illegal Tags % 14
       (bra!.s tggcode)                           % Code         % 15
       (bra!.s gbadheapitem)                      % Illegal Tags % 16 -16
       (bra!.s gbadheapitem)                      % Illegal Tags % 17 -15
       (bra!.s gbadheapitem)                      % Illegal Tags % 18 -14
       (bra!.s gbadheapitem)                      % Illegal Tags % 19 -13
       (bra!.s gbadheapitem)                      % Illegal Tags % 20 -12
       (bra!.s gbadheapitem)                      % Illegal Tags % 21 -11
       (bra!.s gbadheapitem)                      % Illegal Tags % 22 -10
                                                  % 23  -9 Lowest Marked Tag
  
TGGHBYTES
       (move!.l (indirect (reg a0)) (reg d1))
       (addq!.l 5 (reg d1))
       (addq!.l 4 (reg d1))
       (and!.l 16777212 (reg d1))
       (bra!.s endgloop1)
  
TGGHHALFWORDS
       (move!.l (indirect (reg a0)) (reg d1))
       (addq!.l 4 (reg d1))
       (and!.l 16777214 (reg d1))
       (add!.l (reg d1) (reg d1))
       (bra!.s endgloop1)
  
TGGHVECT
  
TGGHWRDS
       (move!.l (indirect (reg a0)) (reg d1))
       (addq!.l 2 (reg d1))
       (and!.l (reg d2) (reg d1))                 % Mask with 16#00FFFFFF
       (add!.l (reg d1) (reg d1))
       (add!.l (reg d1) (reg d1))
       (bra!.s endgloop1)
  
GBADHEAPITEM                                      % illegal heap item assumed to be part of a pair
  
TGGUNBOUND
  
TGGID
  
TGGNEGINT
  
TGGPOSINT
  
TGGFIXN
  
TGGBIGN
  
TGGFLTN
  
TGGSTR
  
TGGBYTES                                          % all of these are assumed to be the 1st element
  
TGGHALFWORDS                                      %          of a pair
  
TGGWRDS
  
TGGVECT
  
TGGPAIR
  
TGGEVECT
  
TGGCODE(moveq 8 (reg d1))                         % skip over pair
  
ENDGLOOP1
       (add!.l (reg d1) (reg d7))                 % Accumulate contiguous hole size
       (add!.l (reg d1) (reg d4))                 % Accumulate total garbage bytes
       (adda!.l (reg d1) (reg a0))                % Advance pointer
       (cmpa!.l (reg a2) (reg a0))                % Test for end of scan
       (bge done)
       (move!.b (indirect (reg a0)) (reg d3))
       (cmp!.b (reg d0) (reg d3))                 % Lowest marked tag is 23
       (blt moregarbage)                          %  (23 = -9 + 32)
       (asr!.l 2 (reg d7))                        % Size of hole in words
       (add!.l (reg d7) (reg d6))                 % Update relative offset
       (cmp!.l (reg d5) (reg d6))                 % Check for spillover ( >= 8 )
       (bge!.s nextseg)                           %  >= 8 means crossed segment boundary
       (moveq 128 (reg d1))                       % bit mask 10000000
       (subq 1 (reg d7))
       (asr!.b (reg d7) (reg d1))                 % make mask of Holesize bits
       (rol!.b (reg d6) (reg d1))                 % rotate mask to corresponding place
       (or!.b (indirect (reg a4)) (reg d1))       % include bits in cumulative map
       (move!.l (reg d4) (indirect (reg a4)))     % update Garbage count in table
       (move!.b (reg d1) (indirect (reg a4)))     % put back bit map
       (bra!.s nongarbageloop)
  
COPYNONGARBAGE
       (cmp!.l (reg d5) (reg d6))                 % Check for segment spillover ( >= 8 )
       (blt!.s nongarbageloop)
  
NEXTSEG(move!.l (reg d6) (reg d1))                % Compute Offset spillover
       (and!.w -8 (reg d1))                       % Offset - (Offset MOD 8)
       (eor!.l (reg d1) (reg d6))                 % Offset MOD 8
       (asr!.l 1 (reg d1))                        % Offset DIV 8 times 4
       (adda!.l (reg d1) (reg a4))
       (move!.l (reg d4) (indirect (reg a4)))     % Total Garbage to Relocation
  
NONGARBAGELOOP
       (ext!.w (reg d3))                          % sign extend to 16 bits
       (sub!.w (reg a6) (reg d3))                 % Clear marktag (subtract 32)
       (move!.b (reg d3) (indirect (reg a0)))     % Restore unmarked tag to heap
       (cmp!.w (reg a5) (reg d3))                 % Test Upper Bound (15)
       (bgt!.s badngitem)
  
NGCASE (add!.w (reg d3) (reg d3))                 % multiply by 2 to index into jump table
                                                  %% (jmp (indexed (reg D3) (displacement 0) jumptable))
       (fullword 1325084692)                      % case jump on tag of item
       (bra!.s tnghbytes)                         % HBytes       % -9
       (bra!.s tnghhalfwords)                     % HHalfWords   % -8
       (bra!.s tnghwrds)                          % HWrds        % -7
       (bra!.s tnghvect)                          % HVect        % -6
       (bra!.s badngitem)                         % Forward      % -5     items illegal in heap
       (bra!.s badngitem)                         % BtrTag       % -4     items illegal in heap
       (bra!.s tngunbound)                        % Unbound      % -3
       (bra!.s tngid)                             % ID           % -2
       (bra!.s tngnegint)                         % NegInt       % -1
  
NGJUMPTABLE
       (bra!.s tngposint)                         % PosInt       % 0
       (bra!.s tngfixn)                           % FixN         % 1
       (bra!.s tngbign)                           % BigN         % 2
       (bra!.s tngfltn)                           % FltN         % 3
       (bra!.s tngstr)                            % Str          % 4
       (bra!.s tngbytes)                          % Bytes        % 5
       (bra!.s tnghalfwords)                      % HalfWords    % 6
       (bra!.s tngwrds)                           % Wrds         % 7
       (bra!.s tngvect)                           % Vect         % 8
       (bra!.s tngpair)                           % Pair         % 9
       (bra!.s tngevect)                          % Evect        % 10
       (bra!.s badngitem)                         % Illegal Tags % 11
       (bra!.s badngitem)                         % Illegal Tags % 12
       (bra!.s badngitem)                         % Illegal Tags % 13
       (bra!.s badngitem)                         % Illegal Tags % 14
       (bra!.s tngcode)                           % Code         % 15
  
TNGHBYTES
       (move!.l (indirect (reg a0)) (reg d1))
       (addq!.l 5 (reg d1))
       (addq!.l 4 (reg d1))
       (and!.l (reg d2) (reg d1))                 % Mask with 16#00FFFFFF
       (asr!.l 2 (reg d1))
       (bra!.s endngloop1)
  
TNGHHALFWORDS
       (move!.l (indirect (reg a0)) (reg d1))
       (addq!.l 4 (reg d1))
       (and!.l (reg d2) (reg d1))                 % Mask with 16#00FFFFFF
       (asr!.l 1 (reg d1))
       (bra!.s endngloop1)
  
TNGHVECT
  
TNGHWRDS
       (move!.l (indirect (reg a0)) (reg d1))
       (addq!.l 2 (reg d1))
       (and!.l (reg d2) (reg d1))                 % Mask with 16#00FFFFFF
       (bra!.s endngloop1)
  
BADNGITEM                                         % illegal heap item assumed to be part of a pair
  
TNGUNBOUND
  
TNGID
  
TNGNEGINT
  
TNGPOSINT
  
TNGFIXN
  
TNGBIGN
  
TNGFLTN
  
TNGSTR
  
TNGBYTES                                          % all of these are assumed to be the 1st element
  
TNGHALFWORDS                                      %          of a pair
  
TNGWRDS
  
TNGVECT
  
TNGPAIR
  
TNGEVECT
  
TNGCODE(moveq 2 (reg d1))                         % skip over pair
  
ENDNGLOOP1
       (add!.l (reg d1) (reg d6))                 % update Offset
       (subq!.l 1 (reg d1))                       % loop count, move all good data down
  
MOVELOOP
       (move!.l (postincrement (reg a0)) (postincrement (reg a3)))
       (dbf (reg d1) moveloop)
       (cmpa!.l (reg a2) (reg a0))                % check for end of scan
       (blt gtestloop)
  
DONE   (move!.l (reg d4) (fluid totalgarbagebytes))       %  (movea!.w 0 (reg A6))          % Lisp Reg A6 convention          scs
       (*pop (reg nil))                           % More conventions
       (rts)))

(lap '((*entry updateeachbase expr 0)             % Assembly routine to relocate
       % all pointers
       %  Register usage
       %
       % D0  end of active heap = CKHeapLast      A0  pointer to item being relocated
       % D1  size computation, etc.               A1  active area = CKFirstHolePtr
       % D2  computational temp                   A2  upper bound of traced region
       % D3  tag byte computation                 A3  index through SYMNAM area
       % D4  lowest  legal tag = -9               A4  offset SYMPRP - SYMNAM
       % D5  highest legal tag = 15               A5  offset SYMVAL - SYMNAM
       % D6  mask to zap tag = 16#00FFFFFF        A6  reloc. table = GCArrayLowerBound
       % D7  outer level counter                  A7 = SP
       % Setup
       (*push (reg nil))                          % Save for later             scs
       (movea!.l (fluid gcarraylowerbound) (reg a6))      % Relocation Table
       (movea!.l (fluid ckfirstholeptr) (reg a1))         % Active area lower bound
       (moveq -9 (reg d4))                        % Lowest  Legal Tag
       (moveq 15 (reg d5))                        % Highest Legal Tag
       (move!.l 16777215 (reg d6))                % To Help Stamp Out Tags
       (move!.l (fluid ckheaplast) (reg d0))              % Upper bound of active area
       (movea!.l (fluid heaplowerbound) (reg a0))         % Heap
       (movea!.l (fluid heaplast) (reg a2))
       (bsr endrangeloop)
       (movea!.l (fluid stackstart) (reg a0))             % Execution Stack
       (movea!.l (fluid stackend) (reg a2))
       (bsr endrangeloop)
       (movea!.l (fluid bndstklowerbound) (reg a0))       % Binding Stack
       (movea!.l (fluid bndstkptr) (reg a2))
       (addq 4 (reg a0))                          % Funny Indexing
       (addq 4 (reg a2))
       (bsr endrangeloop)                         % Symbol Table
       (*move (fluid symnam) (reg a3))                      % Pointer into Name Array
       (*move (fluid symprp) (reg a4))                      %              Property List
       (*move (fluid symval) (reg a5))                      %          and Value Cell
       (suba!.l (reg a3) (reg a4))                % Save offsets only for these
       (suba!.l (reg a3) (reg a5))                %
       (move!.w (wconst maxsymbols) (reg d7))     % Count through symbols
       (subq 4 (reg a3))

       LOOP1
       (addq 4 (reg a3))
       (move!.b (indirect (reg a3)) (reg d1))     % Test for Valid ID
       (dbne (reg d7) loop1)
       (beq!.s done)
       (lea (indirect (reg a3)) (reg a0))
       (bsr relocateitem)
       (lea (indexed (reg a4) (displacement (reg a3) 0)) (reg a0))
       (bsr relocateitem)
       (lea (indexed (reg a5) (displacement (reg a3) 0)) (reg a0))
       (bsr relocateitem)
       (dbf (reg d7) loop1)

       DONE                                       % (movea!.w 0 (reg A6))            % Lisp convention                scs
       (*pop (reg nil))                           % This too
       (rts)

       RANGELOOP
       (move!.b (indirect (reg a0)) (reg d3))     % Extract tag as signed 8 bit quantity
       (bge!.s postag)                            % Test for valid (value in range) tag
       (cmp!.b (reg d4) (reg d3))                 % Test Lower Bound
       (bge!.s case)

       TAGFORWARD                                 % Legal only in that other Garbage Collector


       BADHEAPITEM
       (bsr (internalentry gcbaditem))            % report the infraction
       (addq 4 (reg a0))
       (bra!.s endrangeloop)

       POSTAG
       (cmp!.b (reg d5) (reg d3))                 % Test Upper Bound
       (bgt!.s badheapitem)

       CASE
       (ext!.w (reg d3))                          % sign extend to 16 bits
       (add!.w (reg d3) (reg d3))                 % multiply by 2 to index into jump table
                                                  %% (jmp (indexed (reg D3) (displacement 0) jumptable))
       (fullword 1325084692)                      % case jump on tag of item
       (bra!.s taghbytes)                         % HBytes       % -9
       (bra!.s taghhalfwords)                     % HHalfWords   % -8
       (bra!.s taghwrds)                          % HWrds        % -7
       (bra!.s taghvect)                          % HVect        % -6
       (bra!.s tagforward)                        % Forward      % -5
       (bra!.s tagbtrtag)                         % BtrTag       % -4
       (bra!.s tagunbound)                        % Unbound      % -3
       (bra!.s tagid)                             % ID           % -2
       (bra!.s tagnegint)                         % NegInt       % -1


       JUMPTABLE
       (bra!.s tagposint)                         % PosInt       % 0
       (bra!.s tagfixn)                           % FixN         % 1
       (bra!.s tagbign)                           % BigN         % 2
       (bra!.s tagfltn)                           % FltN         % 3
       (bra!.s tagstr)                            % Str          % 4
       (bra!.s tagbytes)                          % Bytes        % 5
       (bra!.s taghalfwords)                      % HalfWords    % 6
       (bra!.s tagwrds)                           % Wrds         % 7
       (bra!.s tagvect)                           % Vect         % 8
       (bra!.s tagpair)                           % Pair         % 9
       (bra!.s tagevect)                          % Evect        % 10
       (bra!.s badheapitem)                       % Illegal Tags % 11
       (bra!.s badheapitem)                       % Illegal Tags % 12
       (bra!.s badheapitem)                       % Illegal Tags % 13
       (bra!.s badheapitem)                       % Illegal Tags % 14
       (bra!.s tagcode)                           % Code         % 15


       TAGHBYTES
       (move!.l (indirect (reg a0)) (reg d1))
       (addq!.l 5 (reg d1))
       (addq!.l 4 (reg d1))
       (and!.l 16777212 (reg d1))
       (adda!.l (reg d1) (reg a0))
       (bra!.s endrangeloop)

       TAGHHALFWORDS
       (move!.l (indirect (reg a0)) (reg d1))
       (addq!.l 4 (reg d1))
       (and!.l 16777214 (reg d1))
       (add!.l (reg d1) (reg d1))
       (adda!.l (reg d1) (reg a0))
       (bra!.s endrangeloop)

       TAGHWRDS
       (move!.l (indirect (reg a0)) (reg d1))
       (addq!.l 2 (reg d1))
       (and!.l (reg d6) (reg d1))                 % Mask with 16#00FFFFFF
       (add!.l (reg d1) (reg d1))
       (add!.l (reg d1) (reg d1))
       (adda!.l (reg d1) (reg a0))
       (bra!.s endrangeloop)

       TAGFIXN

       TAGBIGN

       TAGFLTN

       TAGSTR

       TAGBYTES

       TAGHALFWORDS

       TAGWRDS

       TAGVECT

       TAGPAIR

       TAGEVECT
       (move!.l (indirect (reg a0)) (reg d1))     % Extract the INF field into D1
       (and!.l (reg d6) (reg d1))                 % using  16#00FFFFFF
       (cmp!.l (reg d0) (reg d1))                 % If the item pointed to by D1
       (bge!.s na)                                % is not within the active area
       (sub!.l (reg a1) (reg d1))                 % there is nothing to do.
       (blt!.s na)
       (bsr!.s relocateitem2)

       NA

       TAGID                                      % ID numbers don't need to be updated


       TAGUNBOUND                                 % Unbound's will be seen in the binding stack


       TAGBTRTAG

       TAGPOSINT                                  % Non-pointer items that do not have to be


       TAGNEGINT                                  % examined further.


       TAGCODE                                    % Do not trace code


       TAGHVECT                                   % Vector header does not have to be examined.
       (addq 4 (reg a0))

       ENDRANGELOOP
       (cmpa!.l (reg a2) (reg a0))                % If A0 is already at the end of the
       (blt rangeloop)                            % range, routine is done.
       (rts)

       RELOCATEITEM
       (move!.b (indirect (reg a0)) (reg d3))     % Tag is a signed 8-bit quantity
       (bge!.s postg2)                            % Test for valid (value in range) tag
       (cmp!.b (reg d4) (reg d3))                 % Test Lower Bound
       (bge!.s case2)
%%     (bra!.s (internalentry gcbaditem))         % report the infraction
       (bra    (internalentry gcbaditem))         % report the infraction


       POSTG2
       (cmp!.b (reg d5) (reg d3))                 % Test Upper Bound
%%     (bgt!.s (internalentry gcbaditem))
       (bgt    (internalentry gcbaditem))

       CASE2                                      % Use a Jump Table to dispatch on the TAG stored in D3
       (ext!.w (reg d3))                          % sign extend to 16 bits
       (add!.w (reg d3) (reg d3))                 % Multiply D3 by 2
                                                  %% The "fullword" below can be replaced when the LAP assembler
                                                  %% handles following statement correctly:
                                                  %%    (jmp (indexed (reg D3) (displacement 0) jumptable2))
       (fullword 1325084692)                      % The Jump Table
       (bra!.s badnonheapitem)         % -9  Legal only in HEAP
       (bra!.s badnonheapitem)         % -8  Legal only in HEAP
       (bra!.s badnonheapitem)         % -7  Legal only in HEAP
       (bra!.s badnonheapitem)         % -6  Legal only in HEAP
       (bra!.s badnonheapitem)         % -5  Legal only in HEAP
       (bra!.s tg2btrtag)                         % BtrTag       % -4
       (bra!.s tg2unbound)                        % Unbound      % -3
       (bra!.s tg2id)                             % ID           % -2
       (bra!.s tg2negint)                         % NegInt       % -1


       JUMPTABLE2
       (bra!.s tg2posint)                         % PosInt       % 0
       (bra!.s tg2fixn)                           % FixN         % 1
       (bra!.s tg2bign)                           % BigN         % 2
       (bra!.s tg2fltn)                           % FltN         % 3
       (bra!.s tg2str)                            % Str          % 4
       (bra!.s tg2bytes)                          % Bytes        % 5
       (bra!.s tg2halfwords)                      % HalfWords    % 6
       (bra!.s tg2wrds)                           % Wrds         % 7
       (bra!.s tg2vect)                           % Vect         % 8
       (bra!.s tg2pair)                           % Pair         % 9
       (bra!.s tg2evect)                          % Evect        % 10
       (bra!.s badnonheapitem)         % 11  Illegal Tags
       (bra!.s badnonheapitem)         % 12  Illegal Tags
       (bra!.s badnonheapitem)         % 13  Illegal Tags
       (bra!.s badnonheapitem)         % 14  Illegal Tags
       (bra!.s tg2code)                           % Code         % 15


BADNONHEAPITEM
       (bra (internalentry gcbaditem))


       TG2STR                                     % Pointer types


       TG2BYTES

       TG2HALFWORDS

       TG2FIXN

       TG2BIGN

       TG2FLTN

       TG2WRDS

       TG2VECT

       TG2EVECT

       TG2PAIR
       (move!.l (indirect (reg a0)) (reg d1))     % Extract the INF field into D1
       (and!.l (reg d6) (reg d1))                 % using  16#00FFFFFF
       (cmp!.l (reg d0) (reg d1))                 % If the item pointed to by D1
       (bge!.s rdone)                             % is not within the active area
       (sub!.l (reg a1) (reg d1))                 % there is nothing to do.
       (blt!.s rdone)

       RELOCATEITEM2                              %  Internal entry point, if tag type already known
       (move!.l (reg d1) (reg d2))                % Find DIV and MOD by 8 * 4
       (and!.w -32 (reg d1))                      % Offset - (Offset MOD 32)
       (eor!.l (reg d1) (reg d2))                 % Offset MOD 32
       (asr!.l 3 (reg d1))                        % DIV 8
       (move!.l (indexed (reg d1) (displacement (reg a6) 0)) (reg d3))
       (move!.b (indexed (reg d1) (displacement (reg a6) 0)) (reg d1))
                                                  % bit map
       (beq!.s relocate)
       (and!.l (reg d6) (reg d3))                 % Zap Map with 16#00FFFFFF
       (asr!.w 2 (reg d2))                        % ShiftCount   0 .. 7
       (lsr!.b (reg d2) (reg d1))
       (beq!.s relocate)

       BCOUNT
       (subq!.l 4 (reg d3))                       % Adjust Relocation by Garbage Bits
       (move!.b (reg d1) (reg d2))
       (subq!.b 1 (reg d2))                       % This litte algorithm counts the
       (and!.b (reg d2) (reg d1))                 % bits in D1.  Do you dig it?
       (bne!.s bcount)

       RELOCATE
       (sub!.l (reg d3) (indirect (reg a0)))      % Adjust the pointer item


       TG2ID                                      % ID's do not need to be relocated


       TG2POSINT                                  % Non-pointer items that do not have to be


       TG2NEGINT                                  % examined further.


       TG2CODE

       TG2BTRTAG

       TG2UNBOUND

       RDONE
       (rts)))

% End of UpdateEachBase, Relocate Item, etc.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%       GCBadItem - a routine to record observed illegal LISP items          %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Given the address (in Register A0) of an improperly tagged object,
%   records the address and the value of the most recent instance
%   in variables GCBadAddress and GCBadObject where they can be
%   accessed during debugging

(lap '((*entry gcbaditem expr 0)
       (addq!.l 1 (fluid gcbaditems))             % Increment the count of occurences
                                                  % record the particulars of the most recent bad item:
       (*move (reg a0) (fluid gcbadaddress))    % Address of bad item is
%%%       (*move (indirect (reg a0))  (fluid gcbadtag))
                                                  % The bad LISP object itself
       (rts)))

% Continue processing as usual
(de gcmessage ()
  (errorprintf "*** GC %w: time %d ms" gcknt* oldtime)
  (errorprintf "*** %d recovered, %d stable, %d active, %d free, %d bad"
               (wquotient totalgarbagebytes addressingunitsperitem)
               (wquotient (wdifference firstholeptr heaplowerbound)
			  addressingunitsperitem)
               (wquotient (wdifference heaplast firstholeptr)
			  addressingunitsperitem)
               (wquotient (wdifference heapupperbound heaplast)
			  addressingunitsperitem)
               gcbaditems))

